var debugChordSequence = "";
var eps = 1E-6;

function NoteType(abbreviation, tonalPitchClass) {
	var letters = "C D EF G A B";
	this.midiPitch = letters.indexOf(abbreviation.charAt(0));
	
	for (var i = 1; i < abbreviation.length; i++) {
		switch (abbreviation.charAt(i)) {
		case "#":
			this.midiPitch = (this.midiPitch + 1) % 12;
			break;
		case "b":
			this.midiPitch = (this.midiPitch - 1 + 12) % 12;
			break;
		}
	}

	this.abbreviation = abbreviation;
	this.tonalPitchClass = tonalPitchClass;
	
	this.toString = function () {
		return this.abbreviation;
	};
}

var noteTypes = [];
var chordTypes = new Array(256);

function init() { }

function initImpl() {
	var prev = null;
	
	var lines = [];
	var file = new QFile(pluginPath + "/tpc.txt");
	if (!file.open(QIODevice.ReadOnly)) {
		QMessageBox.information(0, "error", file.errorString());
		return;
	}
	var stream = new QTextStream(file);
	while (!stream.atEnd()) {
		lines.push(stream.readLine().split(/[\s]+/));
	}
	file.close();

	var c = null; // The note C (a. k. a. Do).
	for (var i = 1; i < lines[0].length; i += 2) {
		for (var j = 0; j < lines.length; j++) {
			var current = new NoteType(lines[j][i], +lines[j][i - 1]);
			noteTypes.push(current);
			if (prev) {
				current.fourth = prev;
				prev.fifth = current;
			}
			prev = current;
			
			if (current.abbreviation == "C") {
				c = current;
			}
		}
	}

	noteTypes[0].fourth = noteTypes[noteTypes.length - 1];
	noteTypes[noteTypes.length - 1].fifth = noteTypes[0];
	
	// Set up minor thirds and major sixths by going by the cycle of fifths.
	var majorSixth = c.fifth.fifth.fifth;
	var current = c;
	while (!current.majorSixth) {
		current.majorSixth = majorSixth;
		current.majorSixth.minorThird = current;
		current = current.fifth;
		majorSixth = current.fifth;
	}
	
	// Sort note types by MIDI pitch so that they can be binary searched. 
	noteTypes.sort(function (x, y) {
		if (x.midiPitch == y.midiPitch) {
			return x.abbreviation < y.abbreviation ? -1 : +1;
		}
		else {
			return x.midiPitch < y.midiPitch ? -1 : +1;
		}
	});
	
	noteTypes.tonality = function (noteType) {
		if (!noteType) {
			return this.noteType;
		}
		this.noteType = noteType;
	};
	
	// We set the actual key later in the run() function based on additional
	// info it has (key signature etc.)
	noteTypes.tonality(c);
	
	noteTypes.get = function (midiPitch, prevNoteType, chordType) {
		midiPitch %= 12;
		
		var left = this.noteType;
		var right = this.noteType;
		
		if (prevNoteType) {
			left = right = prevNoteType;
		}
		if (chordType.isMinor()) {
			left = right = left.majorSixth;
		}

		while (true) {
			if (left.midiPitch == midiPitch) {
				return left;
			}
			else if (right.midiPitch == midiPitch) {
				return right;
			}
			
			left = left.fourth;
			right = right.fifth;
		}
		
		return null;
	};
	
	chordTypes.list = new Array();
	chordTypes.add = function (chordType) {
		this[chordType.id] = chordType;
		this.list.push(chordType);
	};
	
	chordTypes.add(new ChordType("", [0, 4, 7], 1));
	chordTypes.add(new ChordType("m", [0, 3, 7], 16));
	chordTypes.add(new ChordType("sus2", [0, 2, 7], 186));
	chordTypes.add(new ChordType("sus4", [0, 5, 7], 192));
}

function removeAllHarmonies() {
	// TODO: Remove all chord names.
}

function ChordType(abbreviation, intervals, id) {
	this.abbreviation = abbreviation;
	this.intervals = intervals;
	this.id = id;
	this.on = new Array(12);
	this.mask = 0;
	for (var i = 0; i < 12; i++) {
		this.on[i] = 0;
	}
	for (var i = 0; i < intervals.length; i++) {
		this.on[intervals[i]] = 1;
		this.mask |= 1 << intervals[i];
	}
	this.toString = function () {
		return this.abbreviation;
	};
	this.isMinor = function () {
		return /m[\d\+]*/.test(this.abbreviation);
	};
	this.equals = function (that) {
		return that && this.id == that.id;
	};
}

function ChordSlider() {
	// TODO: For now, distinguish only Major, minor and sus chords.
	
	var prev = null;
	var prevNoteType = null;
	var prevChordType = null;
	var prevKey = null;
	var queue = new Array();
	var ticks = new Array();
	
	var bestMatchProb = 0;
	var bestMatch = null;
	var bestNoteType = null;
	var bestChordType = null;
	var bestSkip = 0;
	
	this.proceed = function (current, currentNoteType, currentChordType, currentTick) {
		prevNoteType = currentNoteType;
		prevChordType = currentChordType;
		
		prevKey = currentNoteType;
		if (current != null && currentChordType.isMinor()) {
			prevKey = prevKey.majorSixth;
		}
		
		if (prev != null && current != null &&
			prev.base == current.base &&
			prev.id == current.id &&
			prev.root == current.root) {
			
			debugChordSequence += "-\t";
			return {
				chord: null,
				tick: currentTick
			};
		}
		else if (current == null) {
			debugChordSequence += "?\t";
			
			prev = null;
			return null;
		}
		else {
			debugChordSequence += currentNoteType + currentChordType + "\t";
			
			prev = current;
			return {
				chord: current,
				tick: currentTick
			};
		}
	};
	
	this.check = function (chord, base, iSkip, tonic, mask) {
		var right = mask & ((1 << tonic) - 1);
		mask = (right << (12 - tonic)) | (mask >> tonic);
		
		var addRemove = 0;
		var different = 0;
		
		var p = base.mask ^ mask;
		var q = mask;
		for (var i = 0; i < 12; i++) {
			if (p & 1) {
				addRemove++;
			}
			
			if (q & 1) {
				different++;
			}
			
			p >>= 1;
			q >>= 1;
		}
		var nowProb = 0;
		var len = base.intervals.length;
		if (different >= 3 && len - addRemove >= 2) {
			nowProb = (len - addRemove) / len;
		}
		
		if (nowProb > bestMatchProb) {
			bestSkip = iSkip;
			bestMatchProb = nowProb;
			bestMatch = new Harmony(curScore);
			bestChordType = base;
			bestMatch.id = bestChordType.id;
			bestNoteType = noteTypes.get(tonic, prevKey, base);
			bestMatch.root = bestNoteType.tonalPitchClass;
		}
		
		return nowProb;
	};
	
	this.next = function (chord, tick) {
		var n = chord.notes;
		
		bestMatchProb = 0;
		bestMatch = null;
		bestNoteType = null;
		bestChordType = null;
		bestSkip = 0;
		
		var prevMatchProb = 0;
		for (var iChord = 0; iChord < chordTypes.list.length; iChord++) {
			if (!chordTypes.list[iChord]) {
				continue;
			}
			
			for (var iSkip = 0; iSkip <= queue.length; iSkip++) {
				
				var mask = 0;
				// Add notes from previous non-skipped chords.
				for (var i = iSkip; i < queue.length; i++) {
					mask |= queue[i];
				}
				for (var iNote = 0; iNote < n; iNote++) {
					var note = chord.note(iNote).pitch % 12;
					mask |= 1 << note;
				}
				
				for (var iTonic = 0; iTonic < 12; iTonic++) {
					var prob = this.check(chord, chordTypes.list[iChord], iSkip, iTonic, mask);
					if (chordTypes.list[iChord].equals(prevChordType) && prevNoteType != null && prevNoteType.midiPitch % 12 == iTonic) {
						prevMatchProb = Math.max(prevMatchProb, prob);
					}
				}
			}
		}
		
		if (Math.abs(prevMatchProb - bestMatchProb) < eps) {
			bestMatch = prev;
			bestNoteType = prevNoteType;
			bestChordType = prevChordType;
			bestSkip = 0;
		}
		
		// Remove chords/notes that do not sound anymore.
		queue.splice(0, bestSkip);
		ticks.splice(0, bestSkip);
		
		mask = 0;
		for (var iNote = 0; iNote < n; iNote++) {
			var note = chord.note(iNote).pitch % 12;
			mask |= 1 << note;
		}
		
		if (!queue.length || queue[queue.length] != mask) {
			queue.push(mask);
			ticks.push(tick);
		}
		
		print(queue.length);
		
		return this.proceed(bestMatch, bestNoteType, bestChordType, ticks[0]);
	};
}

function addHarmonies() {
	
	// Determine the song key.
	
	var key = noteTypes.tonality(); // C major.
	var signature = curScore.keysig;
	while (signature < 0) {
		key = key.fourth;
		signature++;
	}
	while (signature > 0) {
		key = key.fifth;
		signature--;
	}
	
	// Add chord names.
	
	var lateCursor = new Cursor(curScore);
	var selectionEnd = new Cursor(curScore);
	
	lateCursor.goToSelectionStart();
	selectionEnd.goToSelectionEnd();
	
	// If there is not selection, operate on the entire score.
	
	var startStaff = lateCursor.staff;
	var endStaff = selectionEnd.staff;
	
	var validCursor = function (cursor) {
		return cursor.tick() < selectionEnd.tick();
	};
	
	var allScore = false;
	if (endStaff <= startStaff) {
		startStaff = 0;
		endStaff = curScore.staves;
		allScore = true;
		lateCursor.rewind();
		
		validCursor = function (cursor) {
			return !cursor.eos();
		};
	}
	
	var cursors = new Array(endStaff);
	for (var i = startStaff; i < endStaff; i++) {
		cursors[i] = new Array(4);
		for (var j = 0; j < 4; j++) {
			cursors[i][j] = new Cursor(curScore);
			cursors[i][j].goToSelectionStart();
			if (allScore) {
				cursors[i][j].rewind();
			}
			cursors[i][j].staff = i;
			cursors[i][j].voice = j;
		}
	}

	var slider = new ChordSlider();
	while (true) {
		
		var curChord = null;
		var curTick = null;
		var curCursor = null;
		for (var i = startStaff; i < endStaff; i++) {
			for (var j = 0; j < cursors[i].length; j++) {
				while (validCursor(cursors[i][j]) && (!cursors[i][j].isChord() || !cursors[i][j].chord().notes)) {
					cursors[i][j].next();
				}
				if (validCursor(cursors[i][j]) && cursors[i][j].isChord() && cursors[i][j].chord().notes) {
					if (!curTick || cursors[i][j].tick() < curTick) {
						curCursor = cursors[i][j];
						curTick = cursors[i][j].tick();
						curChord = new Chord(curScore);
					}
					if (cursors[i][j].tick() == curTick) {
						for (var k = 0; k < cursors[i][j].chord().notes; k++) {
							var note = new Note();
							note.pitch = cursors[i][j].chord().note(k).pitch;
							curChord.addNote(note);
						}
					}
				}
			}
		}
		
		if (!curCursor) {
			break;
		}
		
		var chordDescr = slider.next(curChord, curTick);
		// Add chord title to the score.
		if (chordDescr) {
			lateCursor.staff = curCursor.staff;
			lateCursor.voice = curCursor.voice;
			// If we are certain about the current chord.
			if (chordDescr.chord) {
				// Navigate to a correct position.
				while (lateCursor.tick() < chordDescr.tick || (validCursor(lateCursor) && !lateCursor.isChord())) {
					lateCursor.next();
				}
				
				if (validCursor(lateCursor)) {
					lateCursor.chord().addHarmony(chordDescr.chord);
				}
			}
			
			// To ensure that the current chord is played as long as it can be.
			lateCursor.next();
		}
		
		curCursor.next();
		
	}
}

function run() {
	initImpl();

	if (typeof curScore === "undefined") {
		print("No score is open.");
		return;
	}
	
	if (curScore.hasHarmonies) {
		print("We already have chord names filled in, have to overwrite.");
		removeAllHarmonies();
	}
	
	addHarmonies();
	
	print(debugChordSequence);
}

var mscorePlugin = {
	menu: "Plugins.Magic Chords",
	init: init,
	run: run
};

mscorePlugin;
